@@ -145,9 +145,10 @@ main = hspec $ do
 
     describe "index" $ do
         let test name dummy = prop name $
-              \(NonNegative i) (QCM.NonEmpty xs) ->
+              \(NonNegative i') (QCM.NonEmpty xs) ->
                 let seq' = fromListAs xs dummy
                     mx   = index xs (fromIntegral i)
+                    i    = fromIntegral (i' :: Int)
                  in do
                     mx @?= index seq' i
                     case mx of
